import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import scanpy as sc
import anndata as ad
import muon
from muon import atac as ac
import itertools
from math import ceil
from sklearn import metrics
from plotnine import (ggplot, aes, theme_classic, ylim, element_text, theme,
                      geom_point, scale_x_reverse, annotate, geom_bar, 
                      coord_flip, element_blank, labs, geom_tile, 
                      scale_fill_gradient, facet_wrap, 
                      scale_color_manual, scale_color_gradient)

from scmkl.dataframes import (_parse_result_type, get_weights, sort_groups, 
                              format_group_names)


def _get_alpha(alpha: None | float, result: dict, is_multiclass: bool):
    """
    Gets the smallest alpha from a results file. Works for both binary 
    and multiclass results.
    """
    if type(alpha) == float:
        return alpha
    
    if 'Alpha_star' in result.keys():
        return result['Alpha_star']
    
    if is_multiclass:
        classes = list(result['Classes'])
        alpha_list = list(result[classes[0]]['Norms'].keys())
        alpha = np.min(alpha_list)

    else:
        alpha_list = list(result['Norms'].keys())
        alpha = np.min(alpha_list)

    return alpha


def color_alpha_star(alphas, alpha_star, color):
    """
    Takes an array of alphas and returns a list of the same size where 
    each element is `'black'` except where 
    `alpha_star == alphas`, which will be `'gold'`.

    Parameters
    ----------
    alphas : list | tuple | np.ndarray
        The 1D array of alphas.

    alpha_star: float
        The best performing alpha from cross-validation.

    color : str
        The color of all alphas other than `alpha_star`.

    Returns
    -------
    c_array, c_dict : np.ndarray, dict
        `c_array` is the array of colors corresponding to alphas. 
        `c_dict` is the color dict with alphas as keys and color as 
        values.
    """
    c_array = np.array([color] * len(alphas), dtype='<U15')
    as_pos = np.where(alphas == alpha_star)[0]
    c_array[as_pos] = 'gold'

    c_dict = {alphas[i] : c_array[i]
              for i in range(len(alphas))}

    return c_array, c_dict



def plot_conf_mat(results, title = '', cmap = None, normalize = True,
                          alpha = None, save = None) -> None:
    """
    Creates a confusion matrix from the output of scMKL.

    Parameters
    ----------
    results : dict
        The output from either scmkl.run() or scmkl.one_v_rest()
        containing results from scMKL.

    title : str
        The text to display at the top of the matrix.

    cmap : matplotlib.colors.LinearSegmentedColormap
        The gradient of the values displayed from `matplotlib.pyplot`.
        If `None`, `'Purples'` is used see matplotlib color map 
        reference for more information. 

    normalize : bool
        If `False`, plot the raw numbers. If `True`, plot the 
        proportions.

    alpha : None | float
        Alpha that matrix should be created for. If `results` is from
        `scmkl.one_v_all()`, this is ignored. If `None`, smallest alpha
        will be used.

    save : None | str
        File path to save plot. If `None`, plot is not saved.

    Returns
    -------
    None
    
    Examples
    --------
    >>> # Running scmkl and capturing results
    >>> results = scmkl.run(adata = adata, alpha_list = alpha_list)
    >>> 
    >>> from matplotlib.pyplot import get_cmap
    >>> 
    >>> scmkl.plot_conf_mat(results, title = '', cmap = get_cmap('Blues'))

    ![conf_mat](../tests/figures/plot_conf_mat_binary.png)

    Citiation
    ---------
    http://scikit-learn.org/stable/auto_examples/model_selection/
    plot_confusion_matrix.html
    """
    # Determining type of results
    if ('Observed' in results.keys()) and ('Metrics' in results.keys()):
        multi_class = False
        names = np.unique(results['Observed'])
    else:
        multi_class = True
        names = np.unique(results['Truth_labels'])

    if multi_class:
        cm = metrics.confusion_matrix(y_true = results['Truth_labels'], 
                              y_pred = results['Predicted_class'], 
                              labels = names)
    else:
        min_alpha = np.min(list(results['Metrics'].keys()))
        alpha = alpha if alpha != None else min_alpha
        cm = metrics.confusion_matrix(y_true = results['Observed'],
                              y_pred = results['Predictions'][alpha],
                              labels = names)

    accuracy = np.trace(cm) / float(np.sum(cm))
    misclass = 1 - accuracy

    if cmap is None:
        cmap = plt.get_cmap('Purples')

    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

    plt.figure(figsize=(8, 6))
    plt.imshow(cm, interpolation='nearest', cmap=cmap)
    plt.title(title)
    plt.colorbar()

    tick_marks = np.arange(len(names))
    plt.xticks(tick_marks, names, rotation=45)
    plt.yticks(tick_marks, names)


    thresh = cm.max() / 1.5 if normalize else cm.max() / 2
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        if normalize:
            plt.text(j, i, "{:0.4f}".format(cm[i, j]),
                     horizontalalignment="center",
                     color="white" if cm[i, j] > thresh else "black")
        else:
            plt.text(j, i, "{:,}".format(cm[i, j]),
                     horizontalalignment="center",
                     color="white" if cm[i, j] > thresh else "black")

    acc_label = 'Predicted label\naccuracy={:0.4f}; misclass={:0.4f}'
    acc_label = acc_label.format(accuracy, misclass)

    plt.tight_layout()
    plt.ylabel('True label')
    plt.xlabel(acc_label)

    if save != None:
        plt.savefig(save)
        plt.clf()
    else:
        plt.show()

    return None


def plot_metric(summary_df : pd.DataFrame, alpha_star = None, 
                x_axis: str='Alpha', color = 'red'):
    """
    Takes a data frame of model metrics and optionally alpha star and
    creates a scatter plot given metrics against alpha values. For 
    multiclass results, `alpha_star` is not shown and points are 
    colored by class.
    
    Parameters
    ----------
    summary_df : pd.DataFrame
        Dataframe created by `scmkl.get_summary()`.

    alpha_star : None | float
        If `not None`, a label will be added for tuned `alpha_star` 
        being optimal model parameter for performance from cross 
        validation on the training data. Can be calculated with 
        `scmkl.optimize_alpha()`. Is ignored if `summary_df` is from a 
        multiclass result.

    x_axis : str
        Must be either `'Alpha'` or `'Number of Selected Groups'`. Is 
        the variable that will be plotted on the x-axis.

    color : str
        Color to make points on plot.

    Returns
    -------
    metric_plot : plotnine.ggplot.ggplot
        A plot with alpha values on x-axis and metric on y-axis.

    Examples
    --------
    >>> results = scmkl.run(adata, alpha_list)
    >>> summary_df = scmkl.get_summary(results)
    >>> metric_plot = plot_metric(results)
    >>>
    >>> metric_plot.save('scMKL_performance.png')

    ![metric_plot](../tests/figures/plot_metric_binary.png)
    """
    # Capturing metric from summary_df
    metric_options = ['AUROC', 'Accuracy', 'F1-Score', 'Precision', 'Recall']
    metric = np.intersect1d(metric_options, summary_df.columns)[0]

    x_axis_ticks = np.unique(summary_df[x_axis])
    summary_df['Alpha Star'] = summary_df['Alpha'] == alpha_star

    if 'Class' in summary_df.columns:
        color_lab = 'Class'
    else:
        c_dict = {
            True : 'gold',
            False : color
            }
        color_lab = 'Alpha Star'

    if np.min(summary_df[metric]) < 0.5:
        min_y = 0
    else:
        min_y = 0.6

    plot = (ggplot(summary_df, aes(x = x_axis, y = metric, color=color_lab)) 
            + geom_point(size=4)
            + theme_classic()
            + ylim(min_y, 1)
            + scale_x_reverse(breaks=x_axis_ticks)
            + theme(
                axis_text=element_text(weight='bold', size=10),
                axis_title=element_text(weight='bold', size=12),
                legend_title=element_text(weight='bold', size=14),
                legend_text=element_text(weight='bold', size=12)
            )
            )
    
    if not 'Class' in  summary_df:
        plot += scale_color_manual(c_dict)
        
    return plot


def weights_barplot(result, n_groups: int=1, alpha: None | float=None, 
                 color: str='red'):
    """
    Plots the top `n_groups` weighted groups for each cell class. Works 
    for a single scmkl result (either multiclass or binary).

    Parameters
    ----------
    result : dict
        The output of `scmkl.run()`.

    n_groups : int
        The number of top groups to plot for each cell class.

    alpha : None | float
        The alpha parameter to create figure for. If `None`, the 
        smallest alpha is used.

    color : str
        The color the bars should be.

    Returns
    -------
    plot : plotnine.ggplot.ggplot
        A barplot of weights.

    Examples
    --------
    >>> result = scmkl.run(adata, alpha_list)
    >>> plot = scmkl.weights_barplot(result)

    ![weights_barplot](../tests/figures/weights_barplot_binary.png)
    """
    is_multi, is_many = _parse_result_type(result)
    assert not is_many, "This function only supports single results"

    alpha = _get_alpha(alpha, result, is_multi)
    df = get_weights(result)

    # Subsetting to only alpha and filtering to top groups
    df = df[df['Alpha'] == alpha]
    if is_multi:
        g_order = set(df['Group'])
        for ct in result['Classes']:
            cor_ct = df['Class'] == ct
            other_ct = ~cor_ct

            temp_df = df.copy()
            temp_df = temp_df[cor_ct]
            temp_df = temp_df.sort_values('Kernel Weight', ascending=False)
            top_groups = temp_df['Group'].iloc[0:n_groups].to_numpy()
            
            cor_groups = np.isin(df['Group'], top_groups)
            filter_array = np.logical_and(cor_ct, cor_groups)

            filter_array = np.logical_or(filter_array, other_ct)

            df = df[filter_array]

        df['Group'] = format_group_names(df['Group'], rm_words = ['Markers'])
            

    else:
        df = df.sort_values('Kernel Weight', ascending=False)
        df = df.iloc[0:n_groups]
        df['Group'] = format_group_names(df['Group'], rm_words = ['Markers']) 
        g_order = sort_groups(df)[::-1]
        df['Group'] = pd.Categorical(df['Group'], g_order)

    plot = (ggplot(df)
            + theme_classic()
            + coord_flip()
            + labs(y=f'Kernel Weight (λ = {alpha})')
            + theme(
                axis_text=element_text(weight='bold', size=10),
                axis_title=element_text(weight='bold', size=12),
                axis_title_y=element_blank()
            )
            )

    # This needs to be reworked for multiclass runs
    if is_multi:
        height = (3*ceil((len(set(df['Class'])) / 3)))
        print(height)
        plot += geom_bar(aes(x='Group', y='Kernel Weight'), 
                         stat='identity', fill=color)
        plot += facet_wrap('Class', scales='free', ncol=3)
        plot += theme(figure_size=(15,height))
    else:
        plot += geom_bar(aes(x='Group', y='Kernel Weight'), 
                         stat='identity', fill=color)
        plot += theme(figure_size=(7, 9))

    return plot


def weights_heatmap(result, n_groups: None | int=None, 
                    class_lab: str | None=None, low: str='white', 
                    high: str='red', alpha: float | None=None,
                    scale_weights: bool=False):
    """
    Plots a heatmap of kernel weights with groups on the y-axis and 
    alpha on the x-axis if binary result. If a multiclass result, one 
    alpha is used per class and the x-axis is class.

    Parameters
    ----------
    result : dict
        The output of `scmkl.run()`.

    n_groups : int
        The number of top groups to plot. Not recommended for 
        multiclass results.

    class_lab : str | None
        For multiclass results, if `not None`, will only plot group 
        weights for `class_lab`.

    low : str
        The color for low kernel weight.

    high : str
        The color for high kernel weight.

    alpha : None | float
        The alpha parameter to create figure for. If `None`, the 
        smallest alpha is used.

    scale_weights : bool
        If `True`, the the kernel weights will be scaled for each group 
        within each class. Ignored if result is from a binary 
        classification.

    Returns
    -------
    plot : plotnine.ggplot.ggplot
        A heatmap of weights.

    Examples
    --------
    >>> result = scmkl.run(adata, alpha_list)
    >>> plot = scmkl.weights_heatmap(result)

    ![weights_heatmap](../tests/figures/weights_heatmap_binary.png)
    """
    is_multi, is_many = _parse_result_type(result)
    assert not is_many, "This function only supports single results"

    if type(class_lab) is str:
        result = result[class_lab]

    df = get_weights(result)
    df['Group'] = format_group_names(df['Group'], ['Markers'])

    # Filtering and sorting values
    sum_df = df.groupby('Group')['Kernel Weight'].sum()
    sum_df = sum_df.reset_index()
    order = sort_groups(sum_df)[::-1]
    df['Group'] = pd.Categorical(df['Group'], categories=order)

    if type(n_groups) is int:
        sum_df = sum_df.sort_values(by='Kernel Weight', ascending=False)
        top_groups = sum_df.iloc[0:n_groups]['Group'].to_numpy()
        df = df[np.isin(df['Group'], top_groups)]
    else:
        n_groups = len(set(df['Group']))

    df['Alpha'] = pd.Categorical(df['Alpha'], np.unique(df['Alpha']))

    if n_groups > 40:
        fig_size = (7,8)
    elif n_groups < 25:
        fig_size = (7,6)
    else: 
        fig_size = (7,8)

    if 'Class' in df.columns:
        alpha = _get_alpha(alpha, result, is_multi)
        df = df[df['Alpha'] == alpha]
        x_lab = 'Class'
    else:
        x_lab = 'Alpha'

    if scale_weights and is_multi:
        max_norms = dict()
        for ct in set(df['Class']):
            g_rows = df['Class'] == ct
            max_norms[ct] = np.max(df[g_rows]['Kernel Weight'])
            scale_cols = ['Class', 'Kernel Weight']

        new = df[scale_cols].apply(lambda x: x[1] / max_norms[x[0]], axis=1)
        df['Kernel Weight'] = new

        l_title = 'Scaled\nKernel Weight'

    else:
        l_title = 'Kernel Weight'

    plot = (ggplot(df, aes(x=x_lab, y='Group', fill='Kernel Weight'))
            + geom_tile(color='black')
            + scale_fill_gradient(high=high, low=low)
            + theme_classic()
            + theme(
                figure_size=fig_size,
                axis_text=element_text(weight='bold', size=10),
                axis_text_x=element_text(rotation=90),
                axis_title=element_text(weight='bold', size=12),
                axis_title_y=element_blank(),
                legend_title=element_text(text=l_title, weight='bold', size=12),
                legend_text=element_text(weight='bold', size=10)
            ))

    return plot


def weights_dotplot(result, n_groups: None | int=None, 
                    class_lab: str | None=None, low: str='white', 
                    high: str='red', alpha: float | None=None, 
                    scale_weights: bool=False):
    """
    Plots a dotplot of kernel weights with groups on the y-axis and 
    alpha on the x-axis if binary result. If a multiclass result, one 
    alpha is used per class and the x-axis is class.

    Parameters
    ----------
    result : dict
        The output of `scmkl.run()`.

    n_groups : int
        The number of top groups to plot. Not recommended for 
        multiclass results.

    class_lab : str | None
        For multiclass results, if `not None`, will only plot group 
        weights for `class_lab`.

    low : str
        The color for low kernel weight.

    high : str
        The color for high kernel weight.

    alpha : None | float
        The alpha parameter to create figure for. If `None`, the 
        smallest alpha is used.

    scale_weights : bool
        If `True`, the the kernel weights will be scaled for each 
        within each class.

    Returns
    -------
    plot : plotnine.ggplot.ggplot
        A barplot of weights.

    Examples
    --------
    >>> result = scmkl.run(adata, alpha_list)
    >>> plot = scmkl.weights_dotplot(result)

    ![weights_dotplot](../tests/figures/weights_dotplot_binary.png)
    """
    is_multi, is_many = _parse_result_type(result)
    assert not is_many, "This function only supports single results"

    if type(class_lab) is str:
        result = result[class_lab]

    df = get_weights(result)
    df['Group'] = format_group_names(df['Group'], ['Markers'])

    # Filtering and sorting values
    sum_df = df.groupby('Group')['Kernel Weight'].sum()
    sum_df = sum_df.reset_index()
    order = sort_groups(sum_df)[::-1]
    df['Group'] = pd.Categorical(df['Group'], categories=order)

    if type(n_groups) is int:
        sum_df = sum_df.sort_values(by='Kernel Weight', ascending=False)
        top_groups = sum_df.iloc[0:n_groups]['Group'].to_numpy()
        df = df[np.isin(df['Group'], top_groups)]
    else:
        n_groups = len(set(df['Group']))

    df['Alpha'] = pd.Categorical(df['Alpha'], np.unique(df['Alpha']))

    if n_groups > 40:
        fig_size = (7,8)
    elif n_groups < 25:
        fig_size = (7,6)
    else: 
        fig_size = (7,8)

    if 'Class' in df.columns:
        alpha = _get_alpha(alpha, result, is_multi)
        df = df[df['Alpha'] == alpha]
        x_lab = 'Class'
    else:
        x_lab = 'Alpha'

    if scale_weights:
        max_norms = dict()
        for ct in set(df['Class']):
            g_rows = df['Class'] == ct
            max_norms[ct] = np.max(df[g_rows]['Kernel Weight'])
            scale_cols = ['Class', 'Kernel Weight']

        new = df[scale_cols].apply(lambda x: x[1] / max_norms[x[0]], axis=1)
        df['Kernel Weight'] = new

        l_title = 'Scaled\nKernel Weight'

    else:
        l_title = 'Kernel Weight'


    plot = (ggplot(df, aes(x=x_lab, y='Group', fill='Kernel Weight', color='Kernel Weight'))
            + geom_point(size=5)
            + scale_fill_gradient(high=high, low=low)
            + scale_color_gradient(high=high, low=low)
            + theme_classic()
            + theme(
                figure_size=fig_size,
                axis_text=element_text(weight='bold', size=10),
                axis_text_x=element_text(rotation=90),
                axis_title=element_text(weight='bold', size=12),
                axis_title_y=element_blank(),
                legend_title=element_text(text=l_title, weight='bold', size=12),
                legend_text=element_text(weight='bold', size=10)
            ))

    return plot


def group_umap(adata: ad.AnnData, g_name: str | list, is_binary: bool=False, 
               labels: None | np.ndarray | list=None, title: str='', 
               save: str=''):
    """
    Uses a scmkl formatted `ad.AnnData` object to show sample 
    separation using scmkl discovered groupings.

    Parameters
    ----------
    adata : ad.AnnData
        A scmkl formatted `ad.AnnData` object with `'group_dict'` in 
        `.uns`.

    g_name : str | list
        The groups who's features should be used to filter `adata`. If 
        is a list, features from multiple groups will be used.
    
    is_binary : bool
        If `True`, data will be processed using `muon` which includes 
        TF-IDF normalization and LSI.

    labels : None | np.ndarray | list
        If `None`, labels in `adata.obs['labels']` will be used to 
        color umap points. Else, provided labels will be used to color 
        points.

    title : str
        The title of the plot.

    save : str
        If provided, plot will be saved using `scanpy`'s `save` 
        argument. Should be the desired file name. Output will be 
        `<cwd>/figures/<save>`.

    Returns
    -------
    None

    Examples
    --------
    >>> adata_fp = 'data/_pbmc_rna.h5ad'
    >>> group_fp = 'data/_RNA_azimuth_pbmc_groupings.pkl'
    >>> adata = scmkl.format_adata(adata_fp, 'celltypes', group_fp, 
    ...                            allow_multiclass=True)
    >>> scmkl.group_umap(adata, 'CD16+ Monocyte Markers')

    ![group_umap](../tests/figures/umap_group_rna.png)
    """
    if list == type(g_name):
        feats = {feature 
                 for group in g_name 
                 for feature in adata.uns['group_dict'][group]}
        feats = np.array(list(feats))
    else:
        feats = np.array(list(adata.uns['group_dict'][g_name]))

    if labels:
        assert len(labels) == adata.shape[0], "`labels` do not match `adata`"
        adata.obs['labels'] = labels

    var_names = adata.var_names.to_numpy()

    col_filter = np.isin(var_names, feats)
    adata = adata[:, col_filter].copy()

    if not is_binary:
        sc.pp.normalize_total(adata)
        sc.pp.log1p(adata)
        sc.tl.pca(adata)
        sc.pp.neighbors(adata)
        sc.tl.umap(adata, random_state=1)

    else:
        ac.pp.tfidf(adata, scale_factor=1e4)
        sc.pp.normalize_total(adata)
        sc.pp.log1p(adata)
        ac.tl.lsi(adata)
        sc.pp.scale(adata)
        sc.tl.pca(adata)
        sc.pp.neighbors(adata, n_neighbors=10, n_pcs=30)
        sc.tl.umap(adata, random_state=1)

    if save:
        sc.pl.umap(adata, title=title, color='labels', save=save, show=False)

    else:
        sc.pl.umap(adata, title=title, color='labels')

    return None



import numpy as np
import anndata as ad

from scmkl.train_model import train_model
from scmkl.test import find_selected_groups


def optimize_sparsity(adata: ad.AnnData, group_size: int | None=None, starting_alpha = 1.9, 
                      increment = 0.2, target = 1, n_iter = 10):
    """
    Iteratively train a grouplasso model and update alpha to find the 
    parameter yielding the desired sparsity.
    
    Parameters
    ----------
    adata : ad.AnnData
        `ad.AnnData` with `'Z_train'` and `'Z_test'` in 
        `adata.uns.keys()`.

    group_size : None | int
        Argument describing how the features are grouped. If `None`, 
        `2 * adata.uns['D']` will be used. For more information see 
        [celer documentation](https://mathurinm.github.io/celer/
        generated/celer.GroupLasso.html).

    starting_alpha : float
        The alpha value to start the search at.
    
    increment : float 
        Amount to adjust alpha by between iterations.
    
    target : int
        The desired number of groups selected by the model.

    n_iter : int
        The maximum number of iterations to run.
            
    Returns
    -------
    sparsity_dict : dict
        Tested alpha as keys and the number of selected groups as 
        the values.
        
    alpha : float
        The alpha value yielding the number of selected groups closest 
        to the target.

    Examples
    --------
    >>> sparcity_dict, alpha = scmkl.optimize_sparsity(adata, 
    ...                                                target = 1)
    >>>
    >>> alpha
    0.01

    See Also
    --------
    celer.GroupLasso : https://mathurinm.github.io/celer/
    """
    assert increment > 0 and increment < starting_alpha, ("Choose a positive "
                                                          "increment less "
                                                          "than alpha")
    assert target > 0 and isinstance(target, int), ("Choose an integer "
                                                    "target number of groups "
                                                     "that is greater than 0")
    assert n_iter > 0 and isinstance(n_iter, int), ("Choose an integer "
                                                    "number of iterations "
                                                    "that is greater than 0")

    if group_size == None:
        group_size = adata.uns['D']*2

    sparsity_dict = {}
    alpha = starting_alpha

    for _ in np.arange(n_iter):
        adata = train_model(adata, group_size, alpha)
        num_selected = len(find_selected_groups(adata))

        sparsity_dict[np.round(alpha, 4)] = num_selected

        if num_selected < target:
            #Decreasing alpha will increase the number of selected pathways
            if alpha - increment in sparsity_dict.keys():
                # Make increment smaller so the model can't go back and forth 
                # between alpha values
                increment/=2
            # Ensures that alpha will never be negative
            alpha = np.max([alpha - increment, 1e-3]) 

        elif num_selected > target:
            if alpha + increment in sparsity_dict.keys():
                increment/=2

            alpha += increment
        elif num_selected == target:
            break

    # Find the alpha that minimizes the difference between target and observed
    # number of selected groups
    spar_idx = np.argmin([np.abs(selected - target) 
                          for selected in sparsity_dict.values()])
    optimal_alpha = list(sparsity_dict.keys())[spar_idx]
    
    return sparsity_dict, optimal_alpha



import pandas as pd
import numpy as np
import re


def _find_overlap(start1 : int, end1 : int, start2 : int, end2 : int) -> bool:
    """
    Function to determine whether two regions on the same chromosome 
    overlap.
    
    Parameters
    ----------
    start1 : int
        The start position for region 1
    
    end1 : int
        The end position for region 1
    
    start2 : int
        The start position for region 2
    
    end2: int
        The end postion for region 2
    
    Returns
    -------
    is_overlapping : bool
        `True` if the regions overlap by 1bp. `False` if the regions 
        do not overlap
    """
    return max(start1, start2) <= min(end1, end2)


def get_tss(row : pd.DataFrame) -> int:
    """
    Takes a row from a DataFrame as a DataFrame with columns 
    ['start', 'end', 'strand'] and returns the transcription start site
    depending on gene strandedness.
    
    Parameters
    ----------
    row : pd.DataFrame 
        A gtf dataframe containing only one row and columns 
        `['start', 'end', 'strand']`.
    
    Returns
    -------
    tss : int
        The transcription start site for row's respective annotation.
    """
    if row.iloc[2] == '+':
        return row.iloc[0]
    
    elif row.iloc[2] == '-':
        return row.iloc[1]
    
    else:
        raise ValueError("Strand symbol must be '+' or '-'")
    

def calc_range(row : pd.DataFrame, len_up : int, len_down : int) -> list:
    """
    Returns an infered promotor region for given annotation range 
    depending on transcription start site and user-defined upstream 
    and downstream adjustments.

    Parameters
    ----------
    row : pd.DataFrame 
        A gtf as a dataframe containing only one row and columns 
        `['tss', 'strand']` where `'tss'` is the transcription start 
        site.
    
    len_up : int 
        Number of base pairs upstream from the transcription 
        start site the promotor range should be adjusted for.

    len_down : int
        Number of base pairs downstream from the transcription start 
        site the promotor range should be adjusted for.
    
    Returns
    -------
    start_end : list
        A 2 element list where the first element is the adjusted start
        position and the second the the adjusted end position.
    """
    if row.iloc[1] == '+':
        start = row.iloc[0] - len_up
        end = row.iloc[0] + len_down

    elif row.iloc[1] == '-':
        start = row.iloc[0] - len_down
        end = row.iloc[0] + len_up

    else:
        raise ValueError('Strand symbol must be "+" or "-"')
    
    return [start, end]


def adjust_regions(gene_anno : pd.DataFrame, len_up : int, len_down : int
                    ) -> pd.DataFrame:
    """
    Takes a GTF file as a pd.DataFrame and adjusts start and end 
    positions to represent promotor regions given user-defined 
    upstream and downstream adjustments relative to the transcription
    start site.

    Parameters
    ----------
    gene_anno : pd.DataFrame 
        A dataframe with columns `['chr', 'start', 'end', 'strand', 
        'gene_name']` created from a GTF file.
    
    len_up : int
        Number of base pairs upstream of the transcription start site 
        the promotor region should be adjusted for.
    
    len_down : int
        Number base pairs downstream of the transcription start site 
        the promotor region should be adjusted for.
    
    Returns
    -------
    gene_anno : pd.DataFrame
        A dataframe where `['start', 'end']` columns represent the 
        start and end positions of inferred promotor regions for each 
        annotation.
    """
    # Subsetting DataFrame to only required data
    required_cols = ['chr', 'start', 'end', 'strand', 'gene_name']
    gene_anno = gene_anno[required_cols]

    # Adding a column that annotates transcription start site based on strand
    tss = gene_anno[['start', 'end', 'strand']].apply(get_tss, axis = 1)
    gene_anno.insert(1, column = 'tss', value = tss)

    # Calculating start and end positions based on the tss
    adj_regions = gene_anno[['tss', 'strand']].apply(lambda row:
                                                     calc_range(
                                                        row = row,
                                                        len_up = len_up,
                                                        len_down = len_down), 
                                                   axis = 1, 
                                                   result_type = 'expand'
                                                   )
    
    gene_anno.loc[:, ['start', 'end']] = adj_regions.values
    
    return gene_anno


def create_region_dicts(gene_anno : pd.DataFrame) -> dict:
    """
    Takes a GTF as a pd.DataFrame and returns dictionaries required for
    region comparisons between gene_annotations and assay features.

    Parameters
    ----------
    gene_anno : pd.DataFrame 
        A dataframe with columns `['chr', 'start', 'end', 'strand', 
        'gene_name']` created from a GTF file.
    
    Returns
    -------
    peak_gene_dict : dict
        Keys are regions and values are genes.
        
    ga_regions : dict
        Chromosomes are keys and regions are values.
    """
    peak_gene_dict = {}
    ga_regions = {}

    for _, anno in gene_anno.iterrows():

        cur_chr = anno['chr']
        cur_start = int(anno['start'])
        cur_end = int(anno['end'])
        cur_name = anno['gene_name']

        peak_gene_dict[(cur_chr, cur_start, cur_end)] = cur_name

        cur_region = np.array([[cur_start, cur_end]], dtype = int)
        if cur_chr in ga_regions.keys():
            ga_regions[cur_chr] = np.concatenate((ga_regions[cur_chr], 
                                                  cur_region), axis = 0)
        
        else:
            ga_regions[cur_chr] = cur_region

    return peak_gene_dict, ga_regions


def create_feature_dict(feature_names : list | set | np.ndarray | pd.Series
                         ) -> dict:
    """
    Takes an array of feature names and returns data formatted for
    feature comparisons.
    
    Returns
    -------
    feature_names : array_like
        Array of region names from single-cell epigenetic data matrix.
    
    Returns
    -------
    feature_dict : dict
        Keys are chromosomes and values are regions.
    """
    feature_dict = {}
    feature_names = [re.split(":|-", peak) for peak in feature_names]
    for peak_set in feature_names:
        if peak_set[0] in feature_dict.keys():
            feature_dict[peak_set[0]] = np.concatenate(
                                                (feature_dict[peak_set[0]],
                                                    np.array([[peak_set[1], 
                                                                peak_set[2]]], 
                                                                dtype = int)), 
                                                                axis = 0)
        else:
            feature_dict[peak_set[0]] = np.array([[peak_set[1], 
                                                   peak_set[2]]], 
                                                   dtype = int)

    return feature_dict


def compare_regions(feature_dict : dict, ga_regions : dict,
                     peak_gene_dict : dict, gene_sets : dict, chr_sep : str
                     ) -> dict:
    """
    Takes features from a single-cell data matrix and regions from
    a gene annotation file to return an region grouping where regions 
    from feature_dict and regions from gene annotations overlap.

    Parameters
    ----------
    feature_dict : dict
        Keys are chromosomes and values are regions. This data should 
        come from a single-cell experiment.

    ga_regions : dict
        Keys are chromosomes and values are regions. This data should 
        come from a gene annotations (gtf) file.

    peak_gene_dict : dict
        Keys are peaks from gene annotation file and values are the 
        gene they are associated with.

    gene_sets : dict
        Keys are gene set names and values are an iterable object of 
        gene names.

    chr_sep : str
        The character that separates the chromosome from the rest of 
        the region in the original feature array.
    
    Returns
    -------
    epi_grouping : dict
        Keys are the names from gene_sets and values are a list of 
        regions from `feature_names` that overlap with promotor regions 
        respective to genes in `gene_sets` (i.e., if region in 
        `feature_names` overlaps with promotor region from a gene in a 
        gene set from `gene_sets`, that region will be added to the new 
        dictionary under the respective gene set name).
    """
    epi_grouping = {group : [] for group in gene_sets.keys()}

    for chrom in feature_dict.keys():
        if chrom not in ga_regions.keys():
            continue
        for region in feature_dict[chrom]:
            for anno in ga_regions[chrom]:
                if _find_overlap(region[0], region[1], anno[0], anno[1]):
                    gene = peak_gene_dict[(chrom, anno[0], anno[1])]
                    for group in gene_sets.keys():
                        if gene in gene_sets[group]:
                            feat_region = ''.join((chrom, chr_sep, 
                                                  str(region[0]), "-", 
                                                  str(region[1])))
                            epi_grouping[group].append(feat_region)

    return epi_grouping


def get_region_groupings(gene_anno : pd.DataFrame, gene_sets : dict, 
                       feature_names : np.ndarray | pd.Series | list | set,
                       len_up : int = 5000, len_down : int = 5000) -> dict:
    """
    Creates a peak set where keys are gene set names from `gene_sets` 
    and values are arrays of features pulled from `feature_names`. 
    Features are added to each peak set given overlap between regions 
    in single-cell data matrix and inferred gene promoter regions in 
    `gene_anno`.

    Parameters
    ----------
    gene_anno : pd.DataFrame
        Gene annotations in GTF format as a pd.DataFrame with columns
        `['chr', 'start', 'end', 'strand', 'gene_name']`.

    gene_sets : dict
        Gene set names as keys and an iterable object of gene names
        as values.

    feature_names : array_like | set
        Feature names corresponding to a single_cell epigenetic data 
        matrix.

    Returns
    -------
    epi_grouping : dict
        Keys are the names from `gene_sets` and values
        are a list of regions from `feature_names` that overlap with 
        promotor regions respective to genes in `gene_sets` (i.e., if 
        region in `feature_names` overlaps with promotor region from a 
        gene in a gene set from `gene_sets`, that region will be added 
        to the new dictionary under the respective gene set name).

    Examples
    --------
    >>> # Reading in a gene set and the peak names from dataset
    >>> gene_sets = np.load("data/RNA_hallmark_groupings.pkl", 
    ...                     allow_pickle = True)
    >>> peaks = np.load("data/MCF7_region_names.npy", 
    ...                 allow_pickle = True)
    >>> 
    >>> # Reading in GTF file
    >>> gtf_path = "data/hg38_subset_protein_coding.annotation.gtf"
    >>> gtf = scmkl.read_gtf(gtf_path, filter_to_coding=True)
    >>>
    >>> region_grouping = scmkl.get_region_groupings(gene_anno = gtf,
    ...                                         gene_sets = gene_sets,
    ...                                         feature_names = peaks)
    >>>
    >>> region_grouping.keys()
    dict_keys(['HALLMARK_TNFA_SIGNALING_VIA_NFKB', ...])
    """
    # Getting a unique set of gene names from gene_sets
    all_genes = {gene for group in gene_sets.keys()
                 for gene in gene_sets[group]}
    
    # Filtering out NaN values
    all_genes = [gene for gene in all_genes if type(gene) != float]

    # Filtering out annotations for genes not present in gene_sets
    gene_anno = gene_anno[np.isin(gene_anno['gene_name'], all_genes)]

    # Adjusting start and end columns to represent promotor regions
    gene_anno = adjust_regions(gene_anno = gene_anno, 
                                len_up = len_up, len_down = len_down)

    # Creating a dictionary from assay features where [chr] : (start, end)
    feature_dict = create_feature_dict(feature_names)

    # Creating data structures from gene_anno for comparing regions
    peak_gene_dict, ga_regions = create_region_dicts(gene_anno)

    # Capturing the separator type used in assay
    chr_sep = ':' if ':' in feature_names[0] else '-'

    epi_groupings = compare_regions(feature_dict = feature_dict,
                                     ga_regions = ga_regions,
                                     peak_gene_dict = peak_gene_dict,
                                     gene_sets = gene_sets,
                                     chr_sep = chr_sep)
    
    return epi_groupings



import anndata as ad
import numpy as np


def _check_adatas(adatas : list, check_uns : bool = False, 
                  check_obs : bool = False) -> None:
    """
    Takes a list of AnnData objects and checks that all training and
    testing indices stored in adata.uns and cell labels stored in 
    adata.obs are the same. Ensures that all objects are of type 
    AnnData.
    Args:
        adatas - a list of AnnData objects where each element has the
            same array of indices for adata.uns['train_indices'] and 
            the same array of indices for adata.uns-'test_indices'].
            Additionally, adata.obs['labels'] must be the same for all
            AnnData objects.
        check_uns - a boolean value. If True, function will ensure that
            'train_indices' and 'test_indices' in adata.uns are the
            same. If False, indices will not be checked.
        check_obs - a boolean value. If True, function will ensure that
            adata.obs['labels'] in each AnnData object are the same.
    Returns:
        Returns None. Will throw an error if not all of the criteria
        listed above are met by adatas.
    """
    for i, adata in enumerate(adatas):
        # Ensuring all elements are type AnnData
        if type(adata) != ad.AnnData: 
            raise TypeError("All elements in adatas should be of type "
                            "anndata.AnnData")
        
        # Ensuring all train/test indices are the same
        if check_uns:
            assert np.array_equal(adatas[0].uns['train_indices'],
                                  adata.uns['train_indices']), ("Train "
                                    "indices across AnnData objects in "
                                    "adatas do not match, ensure "
                                    "adata.uns['train_indices'] are the same "
                                    "for all elements")
                
            assert np.array_equal(adatas[0].uns['test_indices'], 
                                    adata.uns['test_indices']), ("Test "
                                    "indices across AnnData objects in "
                                    "adatas do not match, ensure "
                                    "adata.uns['test_indices'] are the same "
                                    "for all elements")

        # Ensuring all cell labels are the same 
        if check_obs:
            assert np.array_equal(adatas[0].obs['labels'], 
                                    adata.obs['labels']), (
                                    "adata.obs['labels'] are different "
                                    "between AnnData objects")
    
    return None


import numpy as np
import sklearn.metrics as skm
import anndata as ad


def predict(adata: ad.AnnData, metrics: list | None=None,
            return_probs: bool=False):
    """
    Function to return predicted labels and calculate any of AUROC, 
    Accuracy, F1 Score, Precision, Recall for a classification. 

    **If labeled_test flag in `adata` is set to `False`,
    metrics cannot be computed.**
    
    Parameters
    ----------
    adata : ad.AnnData
        Has keys `'model'`, `'Z_train'`, and `'Z_test'` in `adata.uns`.

    metrics : list[str] | None
        Which metrics to calculate on the predicted values. Options
        are `'AUROC'`, `'Accuracy'`, `'F1-Score'`, `'Precision'`, and 
        `'Recall'`. If `None`, all five metrics are calculated.

    return_probs : bool
        If `True`, will return a dictionary with class probabilities.

    Returns
    -------
    y_pred : np.ndarray
        Predicted cell classes.

    metrics_dict : dict
        Contains `'AUROC'`, `'Accuracy'`, `'F1-Score'`, 
        `'Precision'`, and/or `'Recall'` keys depending on metrics 
        argument.

    probs : dict
        If `return_probs` is `True`, will return a dictionary with 
        probabilities for each class in `y_test`.

    Examples
    --------
    >>> adata = scmkl.estimate_sigma(adata)
    >>> adata = scmkl.calculate_z(adata)
    >>> metrics = ['AUROC', 'F1-Score', 'Accuracy', 'Precision', 
    ...            'Recall']
    >>> adata = scmkl.train_model(adata, metrics = metrics)
    >>>
    >>> metrics_dict = scmkl.predict(adata)
    >>> metrics_dict.keys()
    dict_keys(['AUROC', 'Accuracy', 'F1-Score', 'Precision', 'Recall'])
    """
    X_test = adata.uns['Z_test']

    allowed_mets = ['AUROC', 'Accuracy', 'F1-Score', 'Precision', 'Recall']

    # Asserting all input metrics are valid
    if metrics is not None:
        mets_allowed = [metric in allowed_mets for metric in metrics]
        assert all(mets_allowed), ("Unknown metric provided. Must be None, "
                                   f"or one or more of {allowed_mets}")

    # Capturing class labels
    train_idx = adata.uns['train_indices']
    classes = np.unique(adata.obs['labels'].iloc[train_idx].to_numpy())

    # Sigmoid function to force probabilities into [0,1]
    probabilities = 1/(1 + np.exp(-adata.uns['model'].predict(X_test)))

    #Convert numerical probabilities into binary phenotype
    y_pred = np.array(np.repeat(classes[1], X_test.shape[0]), 
                      dtype = 'object')
    y_pred[np.round(probabilities, 0).astype(int) == 1] = classes[0]

    if not adata.uns['labeled_test']:
        if not metrics is None:
            print("WARNING: Cannot calculate classification metrics "
                  "for unlabeled test data")
            metrics = None
    else:
        y_test = adata.obs['labels'].iloc[adata.uns['test_indices']]
        y_test = y_test.to_numpy()
        X_test = adata.uns['Z_test']
        assert X_test.shape[0] == len(y_test), ("X rows and length of y must "
                                                "be equal")

        # Group Lasso requires 'continous' y values need to re-descritize it
        y = np.zeros((len(y_test)))
        y[y_test == classes[0]] = 1

        metric_dict = {}

        if (metrics is None) and (return_probs == False):
            return y_pred
        
        # Calculate and save metrics given in metrics
        p_cl = classes[0]
        if 'AUROC' in metrics:
            fpr, tpr, _ = skm.roc_curve(y, probabilities)
            metric_dict['AUROC'] = skm.auc(fpr, tpr)
        if 'Accuracy' in metrics:
            metric_dict['Accuracy'] = np.mean(y_test == y_pred)
        if 'F1-Score' in metrics:
            metric_dict['F1-Score'] = skm.f1_score(y_test, y_pred, 
                                                   pos_label = p_cl)
        if 'Precision' in metrics:
            metric_dict['Precision'] = skm.precision_score(y_test, y_pred, 
                                                           pos_label = p_cl)
        if 'Recall' in metrics:
            metric_dict['Recall'] = skm.recall_score(y_test, y_pred, 
                                                     pos_label = p_cl)

    if return_probs:
        probs = {classes[0] : probabilities,
                 classes[1] : 1 - probabilities}
        if metrics is not None:
            return y_pred, metric_dict, probs
        else:
            return y_pred, probs
    else:
        if metrics is not None:
            return y_pred, metric_dict
        else:
            return y_pred


def find_selected_groups(adata: ad.AnnData) -> np.ndarray:
    """
    Find feature groups selected by the model during training. If 
    feature weight assigned by the model is non-0, then the group 
    containing that feature is selected.

    Parameters
    ----------
    adata : ad.AnnData
        Has `celer.GroupLasso` object in `adata.uns['model']`.

    Returns
    -------
    selected_groups : np.ndarray
        Array containing the names of the groups with nonzero kernel 
        weights.

    Examples
    --------
    >>> adata = scmkl.estimate_sigma(adata)
    >>> adata = scmkl.calculate_z(adata)
    >>> adata = scmkl.train_model(adata)
    >>>
    >>> selected_groups = scmkl.find_selected_groups(adata)
    >>> selected_groups
    np.ndarray(['HALLMARK_ESTROGEN_RESPONSE_EARLY', 
                'HALLMARK_HYPOXIA'])
    """

    selected_groups = []
    coefficients = adata.uns['model'].coef_
    group_size = adata.uns['model'].get_params()['groups']
    group_names = np.array(list(adata.uns['group_dict'].keys()))

    # Loop over the model weights associated with each group and calculate 
    # the L2 norm
    for i, group in enumerate(group_names):
        if not isinstance(group_size, (list, set, np.ndarray, tuple)):
            group_start = i * group_size
            group_end = (i+1) * group_size - 1
            group_cols = np.arange(group_start, group_end)
            group_norm = np.linalg.norm(coefficients[group_cols])
        else: 
            group_norm = np.linalg.norm(coefficients[group_size[i]])

        # Only include the group if the model weights are > 0 
        if group_norm != 0:
            selected_groups.append(group)

    return np.array(selected_groups)



import numpy as np
import pandas as pd
import anndata as ad
from sklearn.metrics import f1_score
import gc

from scmkl.run import run
from scmkl.calculate_z import calculate_z
from scmkl.multimodal_processing import multimodal_processing
from scmkl._checks import _check_adatas


def _eval_labels(cell_labels: np.ndarray, train_indices: np.ndarray, 
                  test_indices: np.ndarray) -> np.ndarray:
    """
    Takes an array of multiclass cell labels and returns a unique array 
    of cell labels to test for.

    Parameters
    ----------
    cell_labels : np.ndarray
        Cell labels that coorespond to an AnnData object.

    train_indices : np.ndarray
        Indices for the training samples in an AnnData object.
    
    test_indices : np.ndarray
        Indices for the testing samples in an AnnData object.

    remove_labels : bool
        If `True`, models will only be created for cell labels in both 
        the training and test data, if `False`, models will be generated
        for all cell labels in the training data.

    Returns
    -------
    uniq_labels : np.ndarray
        Returns a numpy array of unique cell labels to be iterated 
        through during one versus all setups.
    """
    train_uniq_labels = np.unique(cell_labels[train_indices])
    test_uniq_labels = np.unique(cell_labels[test_indices])

    # Getting only labels in both training and testing sets
    uniq_labels = np.intersect1d(train_uniq_labels, test_uniq_labels)

    # Ensuring that at least one cell type label between the two data
    #   are the same
    cl_intersect = np.intersect1d(train_uniq_labels, test_uniq_labels)
    assert len(cl_intersect) > 0, ("There are no common labels between cells "
                                   "in the training and testing samples")

    return uniq_labels


def get_prob_table(results : dict, alpha: float | dict):
    """
    Takes a results dictionary with class and probabilities keys and 
    returns a table of probabilities for each class and the most 
    probable class for each cell.

    Parameters
    ----------
    results : dict
        A nested dictionary that contains a dictionary for each class 
        containing probabilities for each cell class.

    alpha : float | dict
        A float for which model probabilities should be evaluated 
        for.

    Returns
    -------
    prob_table : pd.DataFrame
        Each column is a cell class and the elements are the
        class probability outputs from the model.

    pred_class : list[str]
        The most probable cell classes respective to the training set 
        cells. 

    low_conf : list[bool]
        A bool list where `True`, sample max probability is less than 
        0.5.
    """
    if isinstance(alpha, float):
        prob_table = {class_ : results[class_]['Probabilities'][alpha][class_]
                    for class_ in results.keys()}
    else:
        prob_table = {class_ : list()
                      for class_ in alpha.keys()}
        for class_ in results.keys():
            cur_alpha = alpha[class_]
            prob_table[class_] = results[class_]['Probabilities'][cur_alpha][class_]

    prob_table = pd.DataFrame(prob_table)

    pred_class = []
    maxes = []

    for i, row in prob_table.iterrows():
        row_max = np.max(row)
        indices = np.where(row == row_max)
        prediction = prob_table.columns[indices]

        if len(prediction) > 1:
            prediction = " and ".join(prediction)
        else:
            prediction = prediction[0]

        pred_class.append(prediction)
        maxes.append(row_max)

    maxes = np.round(maxes, 0)
    low_conf = np.invert(np.array(maxes, dtype = np.bool_))

    return prob_table, pred_class, low_conf


def per_model_summary(results: dict, uniq_labels: np.ndarray | list | tuple, 
                      alpha: float) -> pd.DataFrame:
    """
    Takes the results dictionary from `scmkl.one_v_rest()` and adds a 
    summary dataframe show metrics for each model generated from the 
    runs.

    Parameters
    ----------
    results : dict
        Results from `scmkl.one_v_rest()`.

    uniq_labels : array_like
        Unique cell classes from the runs.

    alpha : float | dict
        The alpha for creating the summary from.

    Returns
    -------
    summary_df : pd.DataFrame
        Dataframe with classes on rows and metrics as cols.
    """
    # Getting metrics availible in results
    if isinstance(alpha, dict):
        alpha_key = list(alpha.keys())[0]
        alpha_key = alpha[alpha_key]
        avail_mets = list(results[uniq_labels[0]]['Metrics'][alpha_key])
    else:
        avail_mets = list(results[uniq_labels[0]]['Metrics'][alpha])

    summary_df = {metric : list()
                  for metric in avail_mets}
    summary_df['Class'] = uniq_labels

    for lab in summary_df['Class']:
        for met in avail_mets:
            if isinstance(alpha, dict):
                cur_alpha = alpha[lab]
            else:
                cur_alpha = alpha

            val = results[lab]['Metrics'][cur_alpha][met]
            summary_df[met].append(val)

    return pd.DataFrame(summary_df)


def get_class_train(train_indices: np.ndarray,
                    cell_labels: np.ndarray | list | pd.Series,
                    seed_obj: np.random._generator.Generator,
                    other_factor = 1.5):
    """
    This function returns a dict with each entry being a set of 
    training indices for each cell class to be used in 
    `scmkl.one_v_rest()`.

    Parameters
    ----------
    train_indices : np.ndarray
        The indices in the `ad.AnnData` object of samples availible to 
        train on.

    cell_labels : array_like
        The identity of all cells in the anndata object.

    seed_obj : np.random._generator.Generator
        The seed object used to randomly sample non-target samples.

    other_factor : float
        The ratio of cells to sample for the other class for each 
        model. For example, if classifying B cells with 100 B cells in 
        training, if `other_factor=1`, 100 cells that are not B cells 
        will be trained on with the B cells.

    Returns
    -------
    train_idx : dict
        Keys are cell classes and values are the train indices to 
        train scmkl that include both target and non-target samples.
    """
    uniq_labels = np.unique(cell_labels)
    train_idx = dict()

    if isinstance(cell_labels, pd.Series):
        cell_labels = cell_labels.to_numpy()
    elif isinstance(cell_labels, list):
        cell_labels = np.array(cell_labels)

    for lab in uniq_labels:
        target_pos = np.where(lab == cell_labels[train_indices])[0]
        overlap = np.isin(target_pos, train_indices)

        target_pos = target_pos[overlap]
        other_pos = np.setdiff1d(train_indices, target_pos)

        if (other_factor*target_pos.shape[0]) <= other_pos.shape[0]:
            n_samples = int(other_factor*target_pos.shape[0])
        else:
            n_samples = other_pos.shape[0]

        other_pos = seed_obj.choice(other_pos, n_samples, False)

        lab_train = np.concatenate([target_pos, other_pos])
        train_idx[lab] = lab_train.copy()

    return train_idx


def one_v_rest(adatas : list | ad.AnnData, names : list, 
               alpha_params : np.ndarray, tfidf : list=None, batches: int=10, 
               batch_size: int=100, train_dict: dict=None, 
               force_balance: bool=False, other_factor: float=1.0)-> dict:
    """
    For each cell class, creates model(s) comparing that class to all 
    others. Then, predicts on the training data using `scmkl.run()`.
    Only labels in both training and testing will be run.

    Parameters
    ----------
    adatas : list[AnnData]
        List of `ad.AnnData` objects created by `create_adata()` 
        where each `ad.AnnData` is one modality and composed of both 
        training and testing samples. Requires that `'train_indices'`
        and `'test_indices'` are the same between all `ad.AnnData`s.

    names : list[str]
        String variables that describe each modality respective to 
        `adatas` for labeling.
        
    alpha_params : np.ndarray | float | dict
        If is `dict`, expects keys to correspond to each unique label 
        with float as key (ideally would be the output of 
        scmkl.optimize_alpha). Else, array of alpha values to create 
        each model with or a float to run with a single alpha.

    tfidf : list[bool]
        If element `i` is `True`, `adatas[i]` will be TF-IDF 
        normalized. If `None`, no views will be TF-IDF normalized.

    batches : int
        The number of batches to use for the distance calculation. 
        This will average the result of `batches` distance calculations 
        of `batch_size` randomly sampled cells. More batches will 
        converge to population distance values at the cost of 
        scalability.

    batch_size : int
        The number of cells to include per batch for distance
        calculations. Higher batch size will converge to population
        distance values at the cost of scalability.
        If `batches*batch_size > num_training_cells`,
        `batch_size` will be reduced to 
        `int(num_training_cells / batches)`.

    force_balance : bool
        If `True`, training sets will be balanced to reduce class label 
        imbalance. Defaults to `False`.

    other_factor : float
        The ratio of cells to sample for the other class for each 
        model. For example, if classifying B cells with 100 B cells in 
        training, if `other_factor=1`, 100 cells that are not B cells 
        will be trained on with the B cells.

    Returns
    -------
    results : dict
        Contains keys for each cell class with results from cell class
        versus all other samples. See `scmkl.run()` for futher details. 
        Will also include a probablilities table with the predictions 
        from each model.

    Examples
    --------
    >>> adata = scmkl.create_adata(X = data_mat, 
    ...                            feature_names = gene_names, 
    ...                            group_dict = group_dict)
    >>>
    >>> results = scmkl.one_v_rest(adatas = [adata], names = ['rna'],
    ...                           alpha_list = np.array([0.05, 0.1]),
    ...                           tfidf = [False])
    >>>
    >>> adata.keys()
    dict_keys(['B cells', 'Monocytes', 'Dendritic cells', ...])
    """
    if isinstance(adatas, ad.AnnData):
        adatas = [adatas]
    if isinstance(tfidf, type(None)):
        tfidf = len(adatas)*[False]

    _check_adatas(adatas, check_obs=True, check_uns=True)

    # Want to retain all original train indices
    train_indices = adatas[0].uns['train_indices'].copy()
    test_indices = adatas[0].uns['test_indices'].copy()

    uniq_labels = _eval_labels(cell_labels = adatas[0].obs['labels'], 
                               train_indices = train_indices,
                               test_indices = test_indices)

    if (len(adatas) == 1) and ('Z_train' not in adatas[0].uns.keys()):
        adata = calculate_z(adatas[0], n_features = 5000, 
                            batches=batches, batch_size=batch_size)
    elif len(adatas) > 1:
        adata = multimodal_processing(adatas=adatas, 
                                      names=names, 
                                      tfidf=tfidf,
                                      batches=batches,
                                      batch_size=batch_size)
    else:
        adata = adatas[0].copy()

    # Preventing multiple copies of adata(s) in memory
    del adatas
    gc.collect()

    # Need obj for capturing results
    results = dict()

    # Capturing cell labels to regenerate at each comparison
    cell_labels = np.array(adata.obs['labels'].copy())

    # Capturing perfect train/test splits for each class
    if train_dict:
        train_idx = train_dict
    else:
        if force_balance:
            train_idx = get_class_train(adata.uns['train_indices'], 
                                        cell_labels, 
                                        adata.uns['seed_obj'],
                                        other_factor)

    for label in uniq_labels:

        print(f"Comparing {label} to other types", flush = True)
        cur_labels = cell_labels.copy()
        cur_labels[cell_labels != label] = 'other'

        # Need cur_label vs rest to run model
        adata.obs['labels'] = cur_labels

        if force_balance or train_dict:
            adata.uns['train_indices'] = train_idx[label]

        # Will only run scMKL with tuned alphas
        if isinstance(alpha_params, dict):
            alpha_list = np.array([alpha_params[label]])
        elif isinstance(alpha_params, float):
            alpha_list = np.array([alpha_params])
        else:
            alpha_list = alpha_params
        
        # Running scMKL
        results[label] = run(adata, alpha_list, return_probs=True)

    # Getting final predictions
    if isinstance(alpha_params, dict):
        alpha = alpha_params
    else:
        alpha = np.min(alpha_params)

    prob_table, pred_class, low_conf = get_prob_table(results, alpha)
    macro_f1 = f1_score(cell_labels[adata.uns['test_indices']], 
                        pred_class, average='macro')

    model_summary = per_model_summary(results, uniq_labels, alpha)

    # Global adata obj will be permanently changed if not reset
    adata.obs['labels'] = cell_labels
    adata.uns['train_indices'] = train_indices

    # Need to document vars, probs, and stats
    results['Per_model_summary'] = model_summary
    results['Classes'] = uniq_labels
    results['Probability_table'] = prob_table
    results['Predicted_class'] = pred_class
    results['Truth_labels'] = cell_labels[adata.uns['test_indices']]
    results['Low_confidence'] = low_conf
    results['Macro_F1-Score'] = macro_f1

    if force_balance:
        results['Training_indices'] = train_idx

    return results


import unittest
import scmkl
import numpy as np
from test_calculate_z import create_test_adata  


def train_scmkl():
    """
    This function simply trains a small scMKL model to check the 
    outputs of scmkl.predict() and scmkl.find_selected_groups().
    """
    adata = create_test_adata()
    group_size = adata.uns['D']*2
    alpha = 0.1

    adata = scmkl.estimate_sigma(adata, batch_size=80)
    adata = scmkl.calculate_z(adata, batch_size=80)
    adata = scmkl.train_model(adata, group_size, alpha)

    return adata


class TestTrainModel(unittest.TestCase):
    """
    This unitest class is used to ensure that the scmkl.train_model() 
    function is working properly.
    """

    def test_train_model(self):
        """
        This function ensures that the expected values are returned by 
        scmkl.train_model() by checking the number of coefficients and 
        checking the min and max values of the coefficients.
        """
        adata = train_scmkl()

        # Calculating group size
        group_size = adata.uns['D'] * 2

        # Calculating theoretical number of coefficients
        num_groups = len(adata.uns['group_dict'])
        cor_n_coefs = group_size * num_groups

        # Capturing number of coefficients
        coefs = adata.uns['model'].coef_
        
        # Ensuring that the GroupLasso model has the right number of coeffs
        self.assertEqual(len(coefs), cor_n_coefs, 
                         ("Generated GroupLasso model does not have the "
                          "correct number of coefficients"))
        
        # Checking that the min and max coeffs are expected values
        self.assertAlmostEqual(np.max(coefs), 1.7115070, places = 4, 
                        msg = "Max coefficient is higher than expected")
        self.assertAlmostEqual(np.min(coefs), -1.364834, places = 4,
                        msg = "Min coefficient is lower than expected")
        

if __name__ == '__main__':
    unittest.main()


import unittest
import scmkl
import numpy as np
from test_calculate_z import create_test_adata


class TestEstimateSigma(unittest.TestCase):
    """
    This unittest class is used to test the scmkl.estimate_sigma() 
    and ensure that it is functioning properly.
    """
    
    def test_estimate_sigma(self):
        """
        To test if estimate sigma is functioning properly within 
        scmkl, sigmas are calculated using the example data and both 
        euclidean and cityblock distance metrics then checking that 
        they are all positive and the mean of the sigmas is what we 
        expect.
        """
        adata = create_test_adata()

        # Capturing euclidean sigmas
        adata = scmkl.estimate_sigma(adata, batch_size=80)
        euclidean_sigmas = adata.uns['sigma'].copy()

        # Capturing cityblock sigmas
        adata.uns['distance_metric'] = 'cityblock'
        adata = scmkl.estimate_sigma(adata, batch_size=80)
        cityblock_sigmas = adata.uns['sigma'].copy()

        # Ensuring all sigmas positive
        self.assertTrue(np.all(euclidean_sigmas > 0), 
                        "Euclidean sigmas are not positive")
        self.assertTrue(np.all(cityblock_sigmas > 0), 
                        "Cityblock sigmas are not postive")
        
        # Checking that each has expected mean values
        self.assertAlmostEqual(np.mean(euclidean_sigmas), 11.460746, 
                               places=4, msg=("Euclidean sigmas mean is "
                                                  "not expected value"))
        self.assertAlmostEqual(np.mean(cityblock_sigmas), 75.320433, 
                               places = 4, msg = ("Cityblock sigmas mean is "
                                                  "not expected value"))


if __name__ == '__main__':
    unittest.main()


import scipy
import anndata as ad
import numpy as np

from scmkl.data_processing import process_data, get_group_mat, sample_cells
from scmkl.tfidf_normalize import tfidf


def get_batches(sample_range: list | np.ndarray, 
                seed_obj: np.random._generator.Generator, 
                batches: int,
                batch_size: int) -> np.ndarray:
    """
    Gets batch indices for estimating sigma.

    Parameters
    ----------
    sample_range : list | np.ndarray
        A 1D array with first element being 0 and last element being 
        (1 - number of samples from X_train).

    seed_obj : np.random._generator.Generator
        Numpy random generator object from `adata.uns['seed_obj']`.

    batches : int
        Number of batches to calculate indices for.

    batch_size : int
        Number of samples in each batch.

    Returns
    -------
    batches_idx : np.ndarray
        A 2D array with each row cooresponding to the sample indices 
        for each batch.
    """
    required_n = batches*batch_size
    train_n = len(sample_range)
    assert required_n <= train_n, (f"{required_n} cells required for "
                                   f"{batches} batches of {batch_size} cells; "
                                   f"only {train_n} cells present")

    # Creating a mat of batch x sample indices for estimating sigma
    batches_idx = np.zeros((batches, batch_size), dtype = np.int16)

    for i in range(batches):
        batches_idx[i] = seed_obj.choice(sample_range, 
                                         batch_size, 
                                         replace = False)
        
        # Removing selected indices from sample options
        rm_indices = np.isin(sample_range, batches_idx[i])
        sample_range = np.delete(sample_range, rm_indices)

    return batches_idx


def batch_sigma(X_train: np.ndarray,
                distance_metric: str,
                batch_idx: np.ndarray) -> float:
    """
    Calculates the kernel width (sigma) for a feature grouping through 
    sample batching.

    Parameters
    ----------
    X_train : np.ndarray
        A 2D numpy array with cells x features with features filtered 
        to features in grouping and sampled cells.

    distance_metric: str
        The pairwise distance metric used to estimate sigma. Must
        be one of the options used in scipy.spatial.distance.cdist.

    batch_idx: np.ndarray
        A 2D array with each row cooresponding to the sample indices 
        for each batch.    

    Returns
    -------
    sigma : float
        The estimated group kernel with for Z projection before 
        adjustments for small kernel width or large groupings.
    """
    # Calculate Distance Matrix with specified metric
    n_batches = batch_idx.shape[0]
    batch_sigmas = np.zeros(n_batches)

    for i in np.arange(n_batches):
        idx = batch_idx[i]
        batch_sigma = scipy.spatial.distance.pdist(X_train[idx,:], 
                                                   distance_metric)
        batch_sigmas[i] = np.mean(batch_sigma)

    sigma = np.mean(batch_sigmas)

    return sigma


def est_group_sigma(adata: ad.AnnData,
                    X_train: np.ndarray,
                    n_group_features: int,
                    n_features: int, 
                    batch_idx: np.ndarray) -> float:
    """
    Processes data and calculates the kernel width (sigma) for a 
    feature grouping through sample batching.

    Parameters
    ----------
    X_train : np.ndarray
        A 2D numpy array with cells x features with features filtered 
        to features in grouping and sampled cells.

    adata : anndata.AnnData
        adata used to derive X_train containing 'seed_obj' in uns 
        attribute.

    n_group_features : int
        Number of features in feature grouping.

    n_features : int
        Maximum number of features to be used in sigma estimation.

    batch_idx
        A 2D array with each row cooresponding to the sample indices 
        for each batch.

    Returns
    -------
    sigma : float
        The estimated group kernel with for Z projection.

    """   
    if adata.uns['tfidf']:
        X_train = tfidf(X_train, mode = 'normalize')

    X_train = process_data(X_train, 
                           scale_data=adata.uns['scale_data'], 
                           transform_data=adata.uns['transform_data'],
                           return_dense=True)

    if scipy.sparse.issparse(X_train):
        X_train = X_train.toarray()
        X_train = np.array(X_train, dtype=np.float32)

    # Calculates mean sigma from all batches
    sigma = batch_sigma(X_train, adata.uns['distance_metric'], batch_idx)

    # sigma = 0 is numerically unusable in later steps
    # Using such a small sigma will result in wide distribution, and 
    # typically a non-predictive Z
    if sigma == 0:
        sigma += 1e-5

    if n_features < n_group_features:
        # Heuristic we calculated to account for fewer features used in 
        # distance calculation
        sigma = sigma * n_group_features / n_features 

    return sigma


def estimate_sigma(adata: ad.AnnData,
                   n_features: int = 5000,
                   batches: int = 10, 
                   batch_size: int = 100) -> ad.AnnData:
    """
    Calculate kernel widths to inform distribution for projection of 
    Fourier Features. Calculates one sigma per group of features.

    Parameters
    ----------
    adata : ad.AnnData
        Created by `create_adata`.
    
    n_features : int
        Number of random features to include when estimating sigma. 
        Will be scaled for the whole pathway set according to a 
        heuristic. Used for scalability.

    batches : int
        The number of batches to use for the distance calculation.
        This will average the result of `batches` distance calculations
        of `batch_size` randomly sampled cells. More batches will converge
        to population distance values at the cost of scalability.

    batch_size : int
        The number of cells to include per batch for distance
        calculations. Higher batch size will converge to population
        distance values at the cost of scalability.
        If `batches` * `batch_size` > # training cells,
        `batch_size` will be reduced to `int(num training cells / 
        batches)`.
        
    Returns
    -------
    adata : ad.AnnData
        Key added `adata.uns['sigma']` with grouping kernel widths.

    Examples
    --------
    >>> adata = scmkl.estimate_sigma(adata)
    >>> adata.uns['sigma']
    array([10.4640895 , 10.82011454,  6.16769438,  9.86156855, ...])
    """
    assert batch_size <= len(adata.uns['train_indices']), ("Batch size must be "
                                                          "smaller than the "
                                                          "training set.")

    if batch_size * batches > len(adata.uns['train_indices']):
        old_batch_size = batch_size
        batch_size = int(len(adata.uns['train_indices'])/batches)
        print("Specified batch size required too many cells for "
                "independent batches. Reduced batch size from "
                f"{old_batch_size} to {batch_size}")

    if batch_size > 2000:
        print("Warning: Batch sizes over 2000 may "
               "result in long run-time.")
        
    # Getting subsample indices
    sample_size = np.min((2000, adata.uns['train_indices'].shape[0]))
    indices = sample_cells(adata.uns['train_indices'], sample_size=sample_size, 
                           seed_obj=adata.uns['seed_obj'])

    # Getting batch indices
    sample_range = np.arange(sample_size)
    batch_idx = get_batches(sample_range, adata.uns['seed_obj'], 
                            batches, batch_size)

    # Loop over every group in group_dict
    sigma_array = np.zeros((len(adata.uns['group_dict'])))
    for m, group_features in enumerate(adata.uns['group_dict'].values()):

        n_group_features = len(group_features)

        # Filtering to only features in grouping using filtered view of adata
        X_train = get_group_mat(adata[indices], n_features=n_features, 
                            group_features=group_features, 
                            n_group_features=n_group_features)
        
        if adata.uns['tfidf']:
            X_train = tfidf(X_train, mode='normalize')

        # Data filtering, and transformation according to given data_type
        # Will remove low variance (< 1e5) features regardless of data_type
        # If scale_data will log scale and z-score the data
        X_train = process_data(X_train=X_train,
                               scale_data=adata.uns['scale_data'], 
                               transform_data=adata.uns['transform_data'],
                               return_dense=True)    

        # Estimating sigma
        sigma = est_group_sigma(adata, X_train, n_group_features, 
                                n_features, batch_idx=batch_idx)

        sigma_array[m] = sigma
    
    adata.uns['sigma'] = sigma_array
        
    return adata


import numpy as np
import celer
import anndata as ad


def train_model(adata: ad.AnnData, group_size: int | None=None, alpha:float=0.9):
    """
    Fit a grouplasso model to the provided data.

    Parameters
    ----------
    adata : ad.AnnData 
        Has `'Z_train'` and `'Z_test'` keys in `.uns.keys()`.

    group_size : None | int
        Argument describing how the features are grouped. If `None`, 
        `2 * adata.uns['D']` will be used. For more information see 
        [celer documentation](https://mathurinm.github.io/celer/
        generated/celer.GroupLasso.html).
            
    alpha : float
        Group Lasso regularization coefficient, is a floating point 
        value controlling model solution sparsity. Must be a positive 
        float. The smaller the value, the more feature groups will be 
        selected in the trained model.
    
    Returns
    -------
    adata : ad.AnnData 
        Trained model accessible with `adata.uns['model']`.

    Examples
    --------
    >>> adata = scmkl.estimate_sigma(adata)
    >>> adata = scmkl.calculate_z(adata)
    >>> metrics = ['AUROC', 'F1-Score', 'Accuracy', 'Precision', 
    ...            'Recall']
    >>> d = scmkl.calculate_d(adata.shape[0])
    >>> group_size = 2 * d
    >>> adata = scmkl.train_model(adata, group_size)
    >>>
    >>> 'model' in adata.uns.keys()
    True

    See Also
    --------
    celer :
        https://mathurinm.github.io/celer/generated/celer.GroupLasso.html
    """
    assert alpha > 0, 'Alpha must be positive'

    if group_size is None:
        group_size = 2*adata.uns['D']

    y_train = adata.obs['labels'].iloc[adata.uns['train_indices']]
    X_train = adata.uns['Z_train'][adata.uns['train_indices']]

    cell_labels = np.unique(y_train)

    # This is a regression algorithm. We need to make the labels 'continuous' 
    # for classification, but they will remain binary. Casts training labels 
    # to array of -1,1
    train_labels = np.ones(y_train.shape)
    train_labels[y_train == cell_labels[1]] = -1

    # Alphamax is a calculation to regularize the effect of alpha across 
    # different data sets
    alphamax = np.max(np.abs(X_train.T.dot(train_labels)))
    alphamax /= X_train.shape[0] 
    alphamax *= alpha

    # Instantiate celer Group Lasso Regression Model Object
    model = celer.GroupLasso(groups = group_size, alpha = alphamax)

    # Fit model using training data
    model.fit(X_train, train_labels.ravel())

    adata.uns['model'] = model
    return adata


"""
.. include:: ../README.md
.. include:: ../example/README.md

----------------------------

## **scMKL Documentation**
"""

__version__ = '0.4.0'

__all__ = ['calculate_z', 
           'calculate_d',
           'create_adata', 
           'data_processing',
           'dataframes',
           'estimate_sigma', 
           'extract_results',
           'find_candidates',
           'format_adata',
           'format_group_names',
           'get_folds',
           'get_gene_groupings',
           'get_metrics',
           'get_region_groupings',
           'get_selection',
           'get_summary', 
           'get_weights', 
           'groups_per_alpha',
           'group_umap',
           'multimodal_processing', 
           'one_v_rest', 
           'optimize_alpha', 
           'optimize_sparsity',
           'parse_metrics',
           'parse_weights',
           'plotting',
           'plot_metric',
           'plot_conf_mat',
           'prepare_fold',
           'projections',
           'process_fold',
           'read_files',
           'read_gtf',
           'run',
           'sort_groups',
           'test',
           'tfidf_normalize',
           'train_model',
           'weights_barplot',
           'weights_dotplot',
           'weights_heatmap'
           ]

from scmkl._checks import *
from scmkl.calculate_z import *
from scmkl.create_adata import *
from scmkl.data_processing import *
from scmkl.dataframes import *
from scmkl.estimate_sigma import *
from scmkl.get_gene_groupings import *
from scmkl.get_region_groupings import *
from scmkl.multimodal_processing import *
from scmkl.one_v_rest import *
from scmkl.optimize_alpha import *
from scmkl.optimize_sparsity import *
from scmkl.plotting import *
from scmkl.projections import *
from scmkl.run import *
from scmkl.test import *
from scmkl.tfidf_normalize import *
from scmkl.train_model import *
from scmkl.projections import *


import unittest
import scmkl
import numpy as np
from test_calculate_z import create_test_adata


class TestRun(unittest.TestCase):
    """
    This unittest class is used to ensure scmkl.run() works for both 
    unimodal and multimodal experiments.
    """

    def test_unimodal_run(self):
        """
        This functions ensures that unimodal tasks will run properly 
        for scmkl.run() by checking the output for the correct keys, 
        that all alphas are accounted for, and all samples are 
        accounted for in each model.
        """
        # Creating adata
        adata = create_test_adata()

        # Estimating sigma and Z matrices
        adata = scmkl.estimate_sigma(adata, batch_size=80)
        adata = scmkl.calculate_z(adata, batch_size=80)

        # Running scMKL
        alpha_list = np.array([0.1, 0.25, 0.4])
        num_alphas = alpha_list.shape[0]
        results = scmkl.run(adata, alpha_list)

        expected_keys = ['Metrics', 'Selected_groups', 'Norms', 
                         'Predictions', 'Observed', 'Test_indices', 
                         'Group_names', 'Models', 'Train_time', 'RAM_usage', 
                         'Probabilities']
        
        # Adding training and testing n
        num_observed = len(results['Observed'])
        num_predicted = [results['Predictions'][alpha].shape[0] 
                         for alpha in results['Predictions'].keys()]
        num_predicted.append(num_observed)
        all_n = np.array(num_predicted, dtype = int)

        # Ensuring all samples are accounted for
        self.assertTrue(np.all(all_n == 200),
                         ("Samples from results are missing from 'Observed' "
                          "or 'Predicted'"))

        # Checking that all required keys are present in results
        self.assertTrue(np.all([name in results.keys() 
                                for name in expected_keys]), 
                        "results of scmkl.run() missing keys")
        
        # Checking that data for all alpha is captured
        self.assertEqual(num_alphas, len(results['Metrics'].keys()),
                         "Alphas are missing in the 'Metrics' dictionary")
        self.assertEqual(num_alphas, len(results['Norms'].keys()),
                         "Alphas are missing in the 'Norms' dictionary")
        self.assertEqual(num_alphas, len(results['Selected_groups'].keys()),
                         ("Alphas are missing in the 'Selected_groups' "
                          "dictionary"))
        self.assertEqual(num_alphas, len(results['Models'].keys()),
                         "Alphas are missing in the 'Models' dictionary")
        
    def test_multimodal_run(self):
        """
        This functions ensures that multimodal tasks will run properly 
        for scmkl.run() by checking the output for the correct keys, 
        that all alphas are accounted for, and all samples are 
        accounted for in each model.
        """
        rna_adata = create_test_adata('RNA')
        atac_adata = create_test_adata('ATAC')

        # Combining adatas
        adatas = [rna_adata, atac_adata]
        names = ['RNA', 'ATAC']
        tfidf_list = [False, False]
        alpha_list = np.array([0.1, 0.25, 0.4])
        num_alphas = alpha_list.shape[0]

        # Combining adatas
        adata = scmkl.multimodal_processing(adatas, names, 
                                            tfidf_list, batch_size=80)

        results = scmkl.run(adata, alpha_list)

        expected_keys = ['Metrics', 'Selected_groups', 'Norms', 
                         'Predictions', 'Observed', 'Test_indices', 
                         'Group_names', 'Models', 'Train_time', 'RAM_usage', 
                         'Probabilities']
        
        # Adding training and testing n
        num_observed = len(results['Observed'])
        num_predicted = [results['Predictions'][alpha].shape[0] 
                         for alpha in results['Predictions'].keys()]
        num_predicted.append(num_observed)
        all_n = np.array(num_predicted, dtype = int)

        # Ensuring all samples are accounted for
        self.assertTrue(np.all(all_n == 200),
                         ("Samples from results are missing from 'Observed' "
                          "or 'Predicted'"))

        # Checking that all required keys are present in results
        self.assertTrue(np.all([name in results.keys() 
                                for name in expected_keys]), 
                        "results of scmkl.run() missing keys")
        
        # Checking that data for all alpha is captured
        self.assertEqual(num_alphas, len(results['Metrics'].keys()),
                         "Alphas are missing in the 'Metrics' dictionary")
        self.assertEqual(num_alphas, len(results['Norms'].keys()),
                         "Alphas are missing in the 'Norms' dictionary")
        self.assertEqual(num_alphas, len(results['Selected_groups'].keys()),
                         ("Alphas are missing in the 'Selected_groups' "
                          "dictionary"))
        self.assertEqual(num_alphas, len(results['Models'].keys()),
                         "Alphas are missing in the 'Models' dictionary")
        

if __name__ == '__main__':
    unittest.main()


import unittest
import scmkl
import numpy as np
from test_create_adata import read_data


def create_test_adata(mod = 'RNA'):
    """
    Create a anndata.AnnData object with default params for scmkl
    to run for testing.
    """
    x, grouping, features, labels = read_data(mod)
    d = scmkl.calculate_d(len(labels))
    scale_data = mod == 'RNA'

    adata = scmkl.create_adata(X = x, feature_names = features, 
                                   cell_labels = labels, D = d,
                                   group_dict = grouping, 
                                   remove_features = False,
                                   scale_data=scale_data,
                                   transform_data=scale_data)

    return adata


class TestCalculateZ(unittest.TestCase):
    """
    This unittest class is used to evaluate whether 
    scmkl.calculate_z() is functioning properly.
    """
    def test_calculate_z(self):
        """
        To check the output of calculate_z, z is calculated on both 
        training and testing data then dimensions are checked. Then 
        the maximum and minimum values are check to ensure that 1) 
        both Z_train and Z_test have the same distribution and the 
        values are what we would expect with the given params and data.
        """
        adata = create_test_adata()
        adata = scmkl.estimate_sigma(adata, batch_size=80)
        adata = scmkl.calculate_z(adata, batch_size=80)

        # Capturing theoretical dimensions of Z matrices
        n_cols = 2 * adata.uns['D'] * len(adata.uns['group_dict'])
        train_n_rows = adata.uns['train_indices'].shape[0]
        test_n_rows = adata.uns['test_indices'].shape[0]

        train_z_dims = (train_n_rows, n_cols)
        test_z_dims = (test_n_rows, n_cols)

        # Ensuring Z dims match what is expected
        self.assertEqual(adata.uns['Z_train'].shape, train_z_dims, 
                         "Z train dims are not correct")
        self.assertEqual(adata.uns['Z_test'].shape, test_z_dims,
                         "Z test dims are not correct")

        # Quick check on Z train distribution
        self.assertAlmostEqual(np.min(adata.uns['Z_train']), -0.099999, 
                               places = 4, msg = ("Z_train min outside of "
                                                  "expected distribution"))
        self.assertAlmostEqual(np.max(adata.uns['Z_train']), 0.0999999, 
                               places = 4, msg = ("Z_train max outside of "
                                                  "expected distribution"))
        self.assertAlmostEqual(np.median(adata.uns['Z_train']), 0.064034,
                               places = 4, msg = ("Z_train median is out of "
                                                  "bounds for expected dist"))

        # Quick check on Z test distribution
        self.assertAlmostEqual(np.min(adata.uns['Z_test']), -0.0999999, 
                               places = 4, msg = ("Z_test min outside of "
                                                  "expected distribution"))
        self.assertAlmostEqual(np.max(adata.uns['Z_test']), 0.0999999, 
                               places = 4, msg = ("Z_test max outside of "
                                                  "expected distribution"))
        self.assertAlmostEqual(np.median(adata.uns['Z_test']), 0.063864,
                               places = 4, msg = ("Z_test median is out of "
                                                  "bounds for expected dist"))

if __name__ == '__main__':
    unittest.main()


import numpy as np
import scipy
import anndata as ad

def tfidf(X: np.ndarray | scipy.sparse._csc.csc_matrix, mode: str='filter'):
    """
    Function to use Term Frequency Inverse Document Frequency (TF-IDF)
    filtering for atac data to find meaningful features.
    
    Parameters
    ----------
    X : np.ndarray | scipy.sparse._csc.csc_matrix
        Data matrix of cell x feature.  Must be a `np.ndarray` or 
        `scipy.sparse` matrix.

    mode : str
        Argument to determine what to return. Must be `'filter'` or 
        `'normalize'`.
    
    Returns
    -------
    tfidf | significant_features : array_like
        Output depends on given `'mode'` parameter:

        `'filter'` (np.ndarray): 
        Which column sums are nonzero (i.e. which features are 
        significant).

        `'normalize'` (np.ndarray | scipy.sparse._csc.csc_matrix): 
        TF-IDF filtered data matrix of the same dimensions as `X`. 
    """
    assert mode in ['filter', 'normalize'], ("mode must be 'filter' or "
                                             "'normalize'.")

    if scipy.sparse.issparse(X):
        tf = scipy.sparse.csc_array(X)
        doc_freq = np.array(np.sum(tf > 0, axis=0)).reshape(-1)
    else:
        tf = np.asarray(X)
        doc_freq = np.sum(X > 0, axis=0)

    idf = np.log1p((1 + X.shape[0])/(1 + doc_freq))
    tfidf = tf*idf

    if mode == 'normalize':
        if scipy.sparse.issparse(tfidf):
            tfidf = scipy.sparse.csc_matrix(tfidf)
        return tfidf
    
    elif mode == 'filter':
        significant_features = np.where(np.sum(tfidf, axis=0) > 0)[0]
        return significant_features
        
def tfidf_train_test(X_train, X_test):
    if scipy.sparse.issparse(X_train):
        tf_train = scipy.sparse.csc_array(X_train)
        tf_test = scipy.sparse.csc_array(X_test)
        doc_freq = np.array(np.sum(X_train > 0, axis=0)).reshape(-1)
    else:
        tf_train = X_train
        tf_test = X_test
        doc_freq = np.sum(X_train > 0, axis=0)

    idf = np.log1p((1 + X_train.shape[0])/(1 + doc_freq))

    tfidf_train = tf_train*idf
    tfidf_test = tf_test*idf

    if scipy.sparse.issparse(tfidf_train):
        tfidf_train = scipy.sparse.csc_matrix(tfidf_train)
        tfidf_test = scipy.sparse.csc_matrix(tfidf_test)
        
    return tfidf_train, tfidf_test


def tfidf_normalize(adata: ad.AnnData, binarize: bool=False):
    """
    Function to TF-IDF normalize the data in an adata object. If any 
    rows are entirely 0, that row and its metadata will be removed from
    the object.

    Parameters
    ----------
    adata : ad.AnnData
        `ad.Anndata` with `.X` to be normalized. If `'train_indices'` 
        and `'test_indices'` in `'adata.uns.keys()'`, normalization 
        will be done separately for the training and testing data. 
        Otherwise, it will calculate it on the entire dataset.

    binarize : bool 
        If `True`, all values in `adata.X` greater than 1 will become 
        1.

    Returns
    -------
    adata : ad.AnnData 
        `adata` with `adata.X` TF-IDF normalized. Will now have the 
        train data stacked on test data, and the indices will be 
        adjusted accordingly.

    Examples
    --------
    >>> adata = scmkl.create_adata(X = data_mat, 
    ...                            feature_names = gene_names, 
    ...                            group_dict = group_dict)
    >>> 
    >>> adata = scmkl.tfidf_normalize(adata)
    """
    X = adata.X.copy()
    row_sums = np.sum(X, axis = 1)
    assert np.all(row_sums > 0), "TFIDF requires all row sums be positive"

    if binarize:
        X[X > 0] = 1

    if 'train_indices' in adata.uns_keys():

        train_indices = adata.uns['train_indices'].copy()
        test_indices = adata.uns['test_indices'].copy()

        # Calculate the train TFIDF matrix on just the training data so it is 
        # not biased by testing data
        tfidf_train = tfidf(X[train_indices,:], mode = 'normalize')

        # Calculate the test TFIDF by calculating it on the train and test 
        # data and index the test data
        tfidf_test = tfidf(X, mode = 'normalize')[test_indices,:]

        # Impossible to add rows back to original location so we need to 
        # stack the matrices to maintain train/test
        if scipy.sparse.issparse(X):
            tfidf_norm = scipy.sparse.vstack((tfidf_train, tfidf_test))
        else:
            tfidf_norm = np.vstack((tfidf_train, tfidf_test))

        # I'm not sure why this reassignment is necessary, but without, the 
        # values will be saved as 0s in adata
        adata.uns['train_indices'] = train_indices
        adata.uns['test_indices'] = test_indices

        combined_indices = np.concatenate((train_indices, test_indices))

        # Anndata indexes by "rownames" not position so we need to rename the 
        # rows to properly index
        adata_index = adata.obs_names[combined_indices].astype(int)
        tfidf_norm = tfidf_norm[np.argsort(adata_index),:]

    else:

        tfidf_norm = tfidf(X, mode = 'normalize')

    adata.X = tfidf_norm.copy()

    return adata


from setuptools import setup, find_packages

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name = 'scmkl',
    version = '0.4.0',
    description = "Single-cell analysis using Multiple Kernel Learning",
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    author = 'Sam Kupp, Ian VanGordon, Cigdem Ak',
    author_email = 'kupp@ohsu.edu, vangordi@ohsu.edu, ak@ohsu.edu',
    url = 'https://github.com/ohsu-cedar-comp-hub/scMKL/tree/main',
    packages = find_packages(),
    python_requires = '>=3.11.1, <3.13',
    install_requires = [
        'wheel==0.41.2',
        'anndata==0.10.8',
        'celer==0.7.3',
        'numpy==1.26.4',
        'pandas==2.2.2',
        'scikit-learn==1.5.1',
        'scipy==1.14.1',
        'numba==0.61.2',
        'plotnine==0.14.3',
        'matplotlib==3.9.3',
        'scanpy==1.11.4',
        'umap-learn==0.5.7',
        'muon==0.1.6',
        'gseapy==1.1.9'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ]
)


import unittest
import scmkl
import numpy as np
from test_calculate_z import create_test_adata


class TestMultimodalProcessing(unittest.TestCase):
    """
    This unittest class is used to ensure 
    scmkl.multimodal_processing() is working properly.
    """

    def test_multimodal_processing(self):
        """
        To test whether or not scmkl.multimodal_processing() is 
        working as expected, two adatas will be created and processed. 
        Then, the new dictionary names will be check along with the 
        length of group dict. Then, the dimensions of the resulting 
        adata is checked to ensure the correct number of samples and 
        combined features are present.  
        """
        # Creating adatas
        rna_adata = create_test_adata('RNA')
        atac_adata = create_test_adata('ATAC')

        # Combining adatas
        adatas = [rna_adata, atac_adata]
        names = ['RNA', 'ATAC']
        tfidf_list = [False, False]

        # Combining adatas
        adata = scmkl.multimodal_processing(adatas, names, 
                                            tfidf_list, 
                                            batch_size=80)

        # Capturing prepended names in group_dict
        prep_names = {gname.split('-')[0] 
                      for gname in adata.uns['group_dict'].keys()}
        
        # Ensuring new group names created properly
        self.assertSetEqual(set(names), prep_names, 
                            "Group names incorrectly modified")

        # Ensuring number of groups in correct
        self.assertEqual(len(adata.uns['group_dict']), 100, 
                         "Incorrect number of groupings in combined adata")
        
        # Ensuring number of samples remains the same
        self.assertEqual(adata.shape[0], 1000, 
                         "Incorrect number of samples in final adata")
        
        # Ensuring features were combined correctly
        self.assertEqual(adata.shape[1], 242768, 
                         "Incorrect number of features in final adata")
        

if __name__ == '__main__':
    unittest.main()


import numpy as np
import numba as nb        


@nb.njit(fastmath = True, cache = True)
def gaussian_trans(X: np.ndarray, sigma: float, 
                   seed_obj: np.random._generator.Generator,
                   D: int):
    """
    Function to sample the projection direction vector W
    for calculating gaussian Random Fourier Features (RFF)
    for X.

    Parameters
    ----------
    X : np.ndarray
        Data matrix X to calculate an approximate kernel of.
    
    sigma : float
        Parameter from data distribution controlling the approximate 
        kernel width.

    seed_obj : np.random._generator.Generator
        Numpy random generator object from `adata.uns['seed_obj']`.

    D : int
        Parameter determining the number of RFF used to approximate 
        the kernel function.

    Returns
    -------
    W : np.ndarray
        Vector defining the direction of the projection of RFF.
    """
    gamma = 1 / ( 2*sigma**2)
    sigma_p = 0.5*np.sqrt(2*gamma)

    W = seed_obj.normal(0, sigma_p, X.shape[1]*D)
    W = W.reshape((X.shape[1]), D)

    return W


@nb.njit(fastmath = True, cache = True)
def laplacian_trans(X: np.ndarray, sigma: float, seed_obj, d: int):
    """
    Function to sample the projection direction vector W
    for calculating laplacian Random Fourier Features (RFF)
    for X.

    Parameters
    ----------
    X : np.ndarray
        Data matrix X to calculate an approximate kernel of.
    
    sigma : float
        Parameter from data distribution controlling the approximate 
        kernel width.

    seed_obj : np.random._generator.Generator
        Numpy random generator object from `adata.uns['seed_obj']`.

    D : int
        Parameter determining the number of RFF used to approximate 
        the kernel function.

    Returns
    -------
    W : np.ndarray
        Vector defining the direction of the projection of RFF.
    """
    gamma = 1 / (2 * sigma)

    W = seed_obj.standard_cauchy(X.shape[1] * d)
    W = gamma * W.reshape((X.shape[1], d))

    return W


@nb.njit(fastmath = True, cache = True)
def cauchy_trans(X: np.ndarray, sigma: float, seed_obj, d: int):
    """
    Function to sample the projection direction vector W
    for calculating cauchy Random Fourier Features (RFF)
    for X.

    Parameters
    ----------
    X : np.ndarray
        Data matrix X to calculate an approximate 
        kernel of.
    
    sigma : float
        Parameter from data distribution controlling the approximate 
        kernel width.

    seed_obj : np.random._generator.Generator
        Numpy random generator object from `adata.uns['seed_obj']`.

    D : int
        Parameter determining the number of RFF used to approximate 
        the kernel function.

    Returns
    -------
    W : np.ndarray
        Vector defining the direction of the projection of RFF.
    """
    gamma = 1 / (2 * sigma ** 2)
    b = 0.5 * np.sqrt(gamma)

    W = seed_obj.laplace(0, b, X.shape[1] * d)
    W = W.reshape((X.shape[1], d))

    return W



import numpy as np
import scipy
import anndata as ad

from scmkl.tfidf_normalize import tfidf_train_test
from scmkl.estimate_sigma import est_group_sigma, get_batches
from scmkl.data_processing import process_data, get_group_mat, sample_cells
from scmkl.projections import gaussian_trans, laplacian_trans, cauchy_trans


def get_z_indices(m, D):
    """
    Takes the number associated with the group as `m` and returns the 
    indices for cos and sin functions to be applied.

    Parameters
    ----------
    m : int
        The chronological number of the group being processed.

    D : int
        The number of dimensions per group.

    Returns
    -------
    cos_idx, sin_idx : np.ndarray, np.ndarray
        The indices for cos and sin projections in overall Z matrix.
    """
    x_idx = np.arange(m*2*D ,(m + 1)*2*D)
    cos_idx = x_idx[:len(x_idx)//2]
    sin_idx = x_idx[len(x_idx)//2:]

    return cos_idx, sin_idx


def calc_groupz(X_train, X_test, adata, D, sigma, proj_func):
    """
    Calculates the Z matrix for grouping.

    Parameters
    ----------
    X_train : np.ndarray
        The filtered data matrix to calculate train Z mat for.
    
    X_test : np.ndarray
        The filtered data matrix to calculate test Z mat for.

    adata : anndata.AnnData 
        AnnData object containing `seed_obj` in `.uns` attribute.

    D : int
        Number of dimensions per grouping.

    sigma : float
        Kernel width for grouping.

    proj_func : function
        The projection direction function to be applied to data.

    Returns
    -------
    train_projections, test_projections : np.ndarray, np.ndarray
        Training and testing Z matrices for group.
    """  
    if scipy.sparse.issparse(X_train):
        X_train = X_train.toarray().astype(np.float32)
        X_test = X_test.toarray().astype(np.float32)

    W = proj_func(X_train, sigma, adata.uns['seed_obj'], D)
    
    train_projection = np.matmul(X_train, W, dtype=np.float32)
    test_projection = np.matmul(X_test, W, dtype=np.float32)

    return train_projection, test_projection


def calculate_z(adata, n_features=5000, batches=10, 
                batch_size=100) -> ad.AnnData:
    """
    Function to calculate Z matrices for all groups in both training 
    and testing data.

    Parameters
    ----------
    adata : ad.AnnData
        created by `scmkl.create_adata()` with `adata.uns.keys()`: 
        `'train_indices'`, and `'test_indices'`. 

    n_features : int
        Number of random feature to use when calculating Z; used for 
        scalability.

    batches : int
        The number of batches to use for the distance calculation.
        This will average the result of `batches` distance calculations
        of `batch_size` randomly sampled cells. More batches will converge
        to population distance values at the cost of scalability.

    batch_size : int
        The number of cells to include per batch for distance
        calculations. Higher batch size will converge to population
        distance values at the cost of scalability.
        If `batches*batch_size > num_training_cells`,
        `batch_size` will be reduced to 
        `int(num_training_cells / batches)`.

    Returns
    -------
    adata : ad.AnnData
        `adata` with Z matrices accessible with `adata.uns['Z_train']` 
        and `adata.uns['Z_test']`.

    Examples
    --------
    >>> adata = scmkl.estimate_sigma(adata)
    >>> adata = scmkl.calculate_z(adata)
    >>> adata.uns.keys()
    dict_keys(['Z_train', 'Z_test', 'sigmas', 'train_indices', 
    'test_indices'])
    """
    # Number of groupings taking from group_dict
    n_pathway = len(adata.uns['group_dict'].keys())
    D = adata.uns['D']

    sq_i_d = np.sqrt(1/D)

    # Capturing training and testing sizes
    train_len = len(adata.uns['train_indices'])
    test_len = len(adata.uns['test_indices'])

    if batch_size * batches > train_len:
        old_batch_size = batch_size
        batch_size = int(train_len/batches)
        print("Specified batch size required too many cells for "
                "independent batches. Reduced batch size from "
                f"{old_batch_size} to {batch_size}")

    if 'sigma' not in adata.uns.keys():
        n_samples = np.min((2000, train_len))
        sample_range = np.arange(n_samples)
        batch_idx = get_batches(sample_range, adata.uns['seed_obj'], 
                                batches=batches, batch_size=batch_size)
        sigma_indices = sample_cells(adata.uns['train_indices'], n_samples, adata.uns['seed_obj'])

    # Create Arrays to store concatenated group Zs
    # Each group of features will have a corresponding entry in each array
    n_cols = 2*adata.uns['D']*n_pathway
    Z_train = np.zeros((train_len, n_cols), dtype=np.float32)
    Z_test = np.zeros((test_len, n_cols), dtype=np.float32)


    # Setting kernel function 
    match adata.uns['kernel_type'].lower():
        case 'gaussian':
            proj_func = gaussian_trans
        case 'laplacian':
            proj_func = laplacian_trans
        case 'cauchy':
            proj_func = cauchy_trans


    # Loop over each of the groups and creating Z for each
    sigma_list = list()
    for m, group_features in enumerate(adata.uns['group_dict'].values()):

        n_group_features = len(group_features)

        X_train, X_test = get_group_mat(adata, n_features, group_features, 
                                        n_group_features, process_test=True)
        
        if adata.uns['tfidf']:
            X_train, X_test = tfidf_train_test(X_train, X_test)

        # Data filtering, and transformation according to given data_type
        # Will remove low variance (< 1e5) features regardless of data_type
        # If scale/transform data depending on .uns values
        X_train, X_test = process_data(X_train=X_train, X_test=X_test, 
                                       scale_data=adata.uns['scale_data'], 
                                       transform_data=adata.uns['transform_data'],
                                       return_dense=True)    

        # Getting sigma
        if 'sigma' in adata.uns.keys():
            sigma = adata.uns['sigma'][m]
        else:
            sigma = est_group_sigma(adata, X_train, n_group_features, 
                                    n_features, batch_idx=batch_idx)
            sigma_list.append(sigma)
            
        assert sigma > 0, "Sigma must be more than 0"
        train_projection, test_projection = calc_groupz(X_train, X_test, 
                                                        adata, D, sigma, 
                                                        proj_func)

        # Store group Z in whole-Z object
        # Preserves order to be able to extract meaningful groups
        cos_idx, sin_idx = get_z_indices(m, D)

        Z_train[0:, cos_idx] = np.cos(train_projection, dtype=np.float32)
        Z_train[0:, sin_idx] = np.sin(train_projection, dtype=np.float32)

        Z_test[0:, cos_idx] = np.cos(test_projection, dtype=np.float32)
        Z_test[0:, sin_idx] = np.sin(test_projection, dtype=np.float32)

    adata.uns['Z_train'] = Z_train*sq_i_d
    adata.uns['Z_test'] = Z_test*sq_i_d

    if 'sigma' not in adata.uns.keys():
        adata.uns['sigma'] = np.array(sigma_list)

    return adata